﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Configuration;
using System.Diagnostics;
using System.Threading;
using System.Security;
using BitMoose.Core.Settings;
using BitMoose.Core.Communication;

namespace BitMoose.Core {

	public enum MiningProcessorState {
		Stopped = 0,
		Started = 1
	}

	public class MiningProcessor {

		private BitMooseSettings Settings = null;
		private Dictionary<string, Process> MinerProcesses = new Dictionary<string, Process>();
		private Dictionary<string, FileStream> MinerLogStreams = new Dictionary<string, FileStream>();
		private Server ComServer = null;
		private MiningProcessorState m_State = MiningProcessorState.Stopped;

		#region " Properties "

		public MiningProcessorState State {
			get { return m_State; }
			protected set { m_State = value; }
		}

		#endregion

		#region " Methods "

		/// <summary>
		/// Starts mining activity
		/// </summary>
		public void Start() {
			try {
				LoadConfiguration();
			} catch ( Exception ex ) {
				Trace.TraceError( ex.Message );
				throw;
			}
			StartLogging();
			StartCommunications();
			Trace.TraceInformation( "Communications started." );
			StartMining();
			Trace.TraceInformation( "Mining started." );
			m_State = MiningProcessorState.Started;
		}

		/// <summary>
		/// Stops all mining and logging activity
		/// </summary>
		public void Stop() {
			StopMining();
			Trace.TraceInformation( "Mining stopped." );
			StopCommunications();
			Trace.TraceInformation( "Communications stopped." );
			StopLogging();
			m_State = MiningProcessorState.Stopped;
		}

		/// <summary>
		/// Loads the configuration from the xml file into the processor
		/// </summary>
		private void LoadConfiguration() {
			string fileName = Path.Combine( Path.GetDirectoryName( Process.GetCurrentProcess().MainModule.FileName ), BitMooseSettings.SettingsFileName );
			this.Settings = new BitMooseSettings();
			this.Settings.Load( fileName );
		}

		#region " Communications "

		private void StartCommunications() {
			ComServer = new Server();
			ComServer.Start();
		}

		private void StopCommunications() {
			if ( ComServer != null ) {
				ComServer.Stop();
				ComServer = null;
			}
		}

		private void WriteToCommunications( Process sender, string data ) {
			if ( sender != null && data != null && sender.StartInfo.EnvironmentVariables.ContainsKey( "BitMoose.MinerName" ) ) {
				if ( ComServer != null ) {
					string name = sender.StartInfo.EnvironmentVariables["BitMoose.MinerName"];
					if ( name != null && Settings.Miners.ContainsKey( name ) && Settings.Miners[name].StatusUpdates ) {
						ComServer.Write( String.Format( "{0}: {1}\0\0", name, data.Trim() ) );
					}
				}
			}
		}

		#endregion

		#region " Mining "

		/// <summary>
		/// Starts all mining processes based on the loaded settings.
		/// </summary>
		private void StartMining() {
			foreach ( var ms in Settings.Miners ) {
				MinerOption option = Settings.MinerOptions[ms.Value.OptionKey];
				MiningPool pool = Settings.Pools.Where( a => a.Name == ms.Value.Pool ).FirstOrDefault();
				string host = ms.Value.Host;
				if ( pool != null && String.IsNullOrEmpty( host ) ) {
					host = pool.Host;
				}
				short port = 8332;
				if ( ms.Value.Port.HasValue ) {
					port = ms.Value.Port.Value;
				} else if ( pool != null ) {
					port = pool.Port;
				}
				string args = Utilities.ParseArguments( option.ArgumentFormat, host, ms.Value.UserName, ms.Value.Password, port, ms.Value.Arguments );
				Process proc = Utilities.CreateMiningProcess(
											ms.Key, option.Path, args, ms.Value.Process.WinDomain,
											ms.Value.Process.WinUserName, ms.Value.Process.WinPassword.ToSecureString(), ms.Value.Process.LoadUserProfile
										);
				proc.StartInfo.RedirectStandardOutput = true;
				proc.StartInfo.RedirectStandardError = true;
				proc.StartInfo.RedirectStandardInput = true;
				proc.EnableRaisingEvents = true;
				proc.OutputDataReceived += new DataReceivedEventHandler( MinerProcess_OutputDataReceived );
				proc.ErrorDataReceived += new DataReceivedEventHandler( MinerProcess_ErrorDataReceived );
				Utilities.StartMiningProcess( proc, Utilities.ToProcessPriority( ms.Value.Process.Priority ), Utilities.ParseCPUAffinity( ms.Value.Process.CPUAffinity ) );
				proc.BeginOutputReadLine();
				proc.BeginErrorReadLine();
				MinerProcesses.Add( ms.Key, proc );
			}
		}

		private void MinerProcess_OutputDataReceived( object sender, DataReceivedEventArgs e ) {
			if ( e.Data.IsNullOrWhiteSpace() == false ) {
				string message = Utilities.CleanString( e.Data );
				if ( Settings.DebugLogging ) {
					Trace.TraceInformation( message );
				}
				WriteToLog( sender as Process, message );
				WriteToCommunications( sender as Process, message );
			}
		}

		private void MinerProcess_ErrorDataReceived( object sender, DataReceivedEventArgs e ) {
			if ( e.Data.IsNullOrWhiteSpace() == false ) {
				string message = Utilities.CleanString( e.Data );
				if ( Settings.DebugLogging ) {
					Trace.TraceError( message );
				}
				WriteToLog( sender as Process, message );
				WriteToCommunications( sender as Process, message );
			}
		}

		/// <summary>
		/// Stops all mining processes by sending a close signal, then after 2.5s forcibly killing the process if it hasn't yet closed.
		/// </summary>
		private void StopMining() {
			foreach ( var p in MinerProcesses ) {
				if ( p.Value != null && p.Value.HasExited == false ) {
					p.Value.CancelErrorRead();
					p.Value.CancelOutputRead();
					p.Value.StandardInput.Write( "\x3" );
					p.Value.StandardInput.Flush();
					p.Value.StandardInput.Close();
					p.Value.CloseMainWindow();
					Trace.TraceInformation( String.Format( "Miner '{0}' process sent close command.", p.Key ) );
				}
			}
			//begin 2.5s worth of spin check to ensure the processes have exited.
			for ( int x = 0; x < 8; x++ ) {
				bool alive = false;
				foreach ( var p in MinerProcesses ) {
					try {
						if ( p.Value != null && p.Value.Responding && p.Value.HasExited == false ) {
							alive = true;
							break;
						}
					} catch ( InvalidOperationException ) { }
				}
				if ( alive ) {
					Thread.Sleep( 500 );
				}
			}
			//kill any process that may still be alive
			foreach ( var p in MinerProcesses ) {
				try {
					if ( p.Value != null && p.Value.HasExited == false ) {
						Trace.TraceWarning( String.Format( "Miner '{0}' process was killed forcibly (didn't terminate after 2.5 seconds).", p.Key ) );
						p.Value.Kill();
						p.Value.Close();
					}
				} catch ( InvalidOperationException ) { }
			}
			MinerProcesses.Clear();
		}

		#endregion

		#region " Logging "

		/// <summary>
		/// Starts application logging
		/// </summary>
		private void StartLogging() {
			if ( Settings.DebugLogging ) {
				TextWriterTraceListener textLog = new TextWriterTraceListener( BitMooseSettings.LogFileName, "BitMoose" );
				textLog.TraceOutputOptions = TraceOptions.DateTime | TraceOptions.ProcessId | TraceOptions.ThreadId;
				textLog.Filter = new EventTypeFilter( SourceLevels.Verbose );
				Trace.Listeners.Add( textLog );
				Trace.AutoFlush = true;
			}
			foreach ( var ms in Settings.Miners ) {
				if ( ms.Value.Logging.Enabled ) {
					string path = Path.Combine( Settings.BitMoosePath, "Miner Log - " + ms.Key + ".log" );
					FileStream fs = new FileStream( path, FileMode.OpenOrCreate, FileAccess.Write, FileShare.Read, 1024 );
					fs.Seek( 0, SeekOrigin.End );
					MinerLogStreams.Add( ms.Key, fs );
				}
			}
		}

		/// <summary>
		/// Stops application logging
		/// </summary>
		private void StopLogging() {
			Trace.Listeners.Clear();
			foreach ( var log in MinerLogStreams ) {
				log.Value.Close();
			}
			MinerLogStreams.Clear();
		}

		/// <summary>
		/// Writes the data for the given process to the log file.
		/// </summary>
		private void WriteToLog( Process sender, string data ) {
			if ( sender != null && data != null && sender.StartInfo.EnvironmentVariables.ContainsKey( "BitMoose.MinerName" ) ) {
				string name = sender.StartInfo.EnvironmentVariables["BitMoose.MinerName"];
				if ( name != null && MinerLogStreams.ContainsKey( name ) && Settings.Miners.ContainsKey( name ) ) {
					FileStream fs = MinerLogStreams[name];
					if ( Settings.Miners[name].Logging.Enabled ) {
						lock ( fs ) {
							if ( fs.Length > Settings.Miners[name].Logging.MaxFileSize ) {
								fs.SetLength( 0 );
							}
							byte[] text = Encoding.UTF8.GetBytes( DateTime.Now.ToString( "MM/dd/yyyy hh:mm:ss tt: " ) + data + "\r\n" );
							fs.Write( text, 0, text.Length );

						}
					} else {
						fs.Close();
						MinerLogStreams.Remove( name );
					}
				}
			}
		}

		#endregion

		#endregion

	}

}
