﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Configuration;
using System.Diagnostics;
using System.IO;
using System.Xml;
using System.Xml.Linq;

namespace BitMoose.Core.Settings {

	public class BitMooseSettings {

		public const string SettingsFileName = "Settings.xml";
		public const string LogFileName = "bitmoose.log";

		private const string DefaultUserNameValue = "";
		private const string DefaultPasswordValue = "";

		#region " Properties "

		public string BitMoosePath { get; protected set; }

		public bool DebugLogging { get; set; }

		public string DefaultUserName { get; set; }

		public string DefaultPassword { get; set; }

		public string DefaultMinerOption { get; set; }

		public string DefaultPool { get; set; }

		public Dictionary<string, MinerOption> MinerOptions { get; protected set; }

		public Dictionary<string, MinerSetting> Miners { get; protected set; }

		public MiningPool[] Pools { get; set; }

		#endregion

		#region " Constructor(s) "

		public BitMooseSettings() {
			this.MinerOptions = new Dictionary<string, MinerOption>();
			this.Miners = new Dictionary<string, MinerSetting>();
		}

		#endregion

		#region " Methods "

		#region " Loading "

		/// <summary>
		/// Loads the settings from the XML file.
		/// </summary>
		public void Load( string fileName ) {
			string localDirectory = Path.GetDirectoryName( fileName );
			if ( Directory.Exists( localDirectory ) ) {
				if ( File.Exists( fileName ) ) {
					Dictionary<string, MinerOption> options = new Dictionary<string, MinerOption>();
					Dictionary<string, MinerSetting> miners = new Dictionary<string, MinerSetting>();
					XDocument xdoc = null;
					try {
						xdoc = XDocument.Load( fileName );
					} catch ( Exception ex ) {
						throw new ApplicationException( "Failed to load settings from XML file. " + ex.Message, ex );
					}
					//load the defaults
					bool debugLog = false;
					if ( xdoc.Root.Element( "DebugLogging" ) != null ) {
						debugLog = Convert.ToBoolean( xdoc.Root.Element( "DebugLogging" ).Value );
					}
					string defUserName = ( from n in xdoc.Root.Descendants( "DefaultUserName" )
										   select n.Value ).FirstOrDefault();
					string defPassword = ( from n in xdoc.Root.Descendants( "DefaultPassword" )
										   select n.Value ).FirstOrDefault();
					string defMinerOpt = ( from n in xdoc.Root.Descendants( "DefaultMinerOption" )
										   select n.Value ).FirstOrDefault();
					string defPool = ( from n in xdoc.Root.Descendants( "DefaultPool" )
									   select n.Value ).FirstOrDefault();
					//load the miner options
					foreach ( var element in xdoc.Root.Element( "MinerOptions" ).Descendants( "MinerOption" ) ) {
						string name = null;
						MinerOption opt = LoadMinerOption( element, out name );
						if ( options.ContainsKey( name ) ) {
							throw new ApplicationException( String.Format( "Duplicate miner option name '{0}' detected. Miner names must be unique.", name ) );
						}
						options.Add( name, opt );
					}
					//load the miners
					int minerIndex = 1;
					foreach ( var element in xdoc.Root.Element( "Miners" ).Descendants( "Miner" ) ) {
						string name = null;
						MinerSetting m = LoadMiner( element, out name );
						name = name ?? String.Format( "Miner #{0}", minerIndex );
						m.UserName = m.UserName ?? defUserName;
						m.Password = m.Password ?? defPassword;
						if ( miners.ContainsKey( name ) ) {
							throw new ApplicationException( String.Format( "Duplicate miner name '{0}' detected. Miner names must be unique.", name ) );
						}
						miners.Add( name, m );
						minerIndex++;
					}
					//load the pools
					List<MiningPool> pools = new List<MiningPool>();
					foreach ( var element in xdoc.Root.Element( "Pools" ).Descendants( "Pool" ) ) {
						MiningPool pool = new MiningPool();
						pool.Name = element.Attribute( "name" ).Value;
						pool.Host = element.Attribute( "host" ).Value;
						pool.Port = Convert.ToInt16( element.Attribute( "port" ).Value );
						if ( element.Attribute( "website" ) != null ) {
							pool.Website = element.Attribute( "website" ).Value;
						}
						pools.Add( pool );
					}
					//validate
					if ( String.IsNullOrEmpty( defMinerOpt ) ) {
						throw new ApplicationException( "The default miner option is not specified in the settings file. It must be named as the value for the 'DefaultMinerOption' element." );
					}
					foreach ( var option in options ) {
						if ( File.Exists( option.Value.Path ) == false ) {
							if ( File.Exists( Path.Combine( localDirectory, option.Value.Path ) ) == false ) {
								throw new ApplicationException( String.Format( "Invalid path for miner option '{0}'. The file was not found or is inaccessible.", option.Key ) );
							} else {
								option.Value.Path = Path.Combine( localDirectory, option.Value.Path ); //update path to use combined.
							}
						}
						if ( String.IsNullOrEmpty( option.Key ) ) {
							throw new ApplicationException( "Invalid name on miner option. All miner options must have a name attribute specified with a unique value." );
						}
					}
					foreach ( var miner in miners ) {
						if ( options.ContainsKey( miner.Value.OptionKey ) == false ) {
							throw new ApplicationException( String.Format( "Invalid option value '{1}' on miner '{0}'. The option value must be a name from the available miner options.", miner.Key, miner.Value.OptionKey ) );
						}
						if ( String.IsNullOrEmpty( miner.Key ) ) {
							throw new ApplicationException( "Invalid name on miner. All miners must have a name attribute specified with a unique value." );
						}
					}
					if ( String.IsNullOrEmpty( defPool ) == false && pools.Any( a => a.Name == defPool ) == false ) {
						throw new ApplicationException( "The specified default pool does not exist." );
					}
					//set values
					this.BitMoosePath = localDirectory;
					this.DebugLogging = debugLog;
					this.DefaultUserName = ( defUserName ?? DefaultUserNameValue ).Trim();
					this.DefaultPassword = ( defPassword ?? DefaultPasswordValue ).Trim();
					this.DefaultMinerOption = defMinerOpt;
					this.DefaultPool = defPool;
					this.MinerOptions = options;
					this.Miners = miners;
					this.Pools = pools.ToArray();
				} else {
					throw new FileNotFoundException( String.Format( "The settings file '{0}' was not found or is inaccessible.", fileName ), SettingsFileName );
				}
			} else {
				throw new DirectoryNotFoundException( String.Format( "The directory path '{0}' was not found or is inaccessible.", localDirectory ) );
			}
		}

		/// <summary>
		/// Loads the miner option from the settings XML into an represenative object
		/// </summary>
		private MinerOption LoadMinerOption( XElement optionElement, out string name ) {
			MinerOption opt = new MinerOption();
			opt.Path = optionElement.Attribute( "path" ).Value;
			opt.ArgumentFormat = optionElement.Attribute( "argumentformat" ).Value;
			if ( optionElement.Attribute( "arguments" ) != null ) {
				opt.Arguments = optionElement.Attribute( "arguments" ).Value;
			}
			name = optionElement.Attribute( "name" ).Value;
			return opt;
		}

		/// <summary>
		/// Loads the miner from the settings XML into an represenative object
		/// </summary>
		private MinerSetting LoadMiner( XElement minerElement, out string name ) {
			MinerSetting m = new MinerSetting();
			name = minerElement.Attribute( "name" ).Value;
			m.OptionKey = minerElement.Attribute( "option" ).Value;
			if ( minerElement.Attribute( "arguments" ) != null ) {
				m.Arguments = minerElement.Attribute( "arguments" ).Value;
			}
			if ( minerElement.Attribute( "pool" ) != null ) {
				m.Pool = minerElement.Attribute( "pool" ).Value;
			}
			if ( minerElement.Attribute( "host" ) != null ) {
				m.Host = minerElement.Attribute( "host" ).Value;
			}
			if ( minerElement.Attribute( "port" ) != null ) {
				short port = 8332;
				if ( short.TryParse( minerElement.Attribute( "port" ).Value, out port ) ) {
					m.Port = port;
				}
			}
			if ( minerElement.Attribute( "username" ) != null ) {
				m.UserName = minerElement.Attribute( "username" ).Value;
			}
			if ( minerElement.Attribute( "password" ) != null ) {
				m.Password = minerElement.Attribute( "password" ).Value;
			}
			if ( minerElement.Attribute( "enabled" ) != null ) {
				m.Enabled = Convert.ToBoolean( minerElement.Attribute( "enabled" ).Value );
			}
			if ( minerElement.Attribute( "netupdate" ) != null ) {
				m.StatusUpdates = Convert.ToBoolean( minerElement.Attribute( "netupdate" ).Value );
			}
			//load logging
			m.Logging = new MinerLogSetting();
			XElement logElement = minerElement.Element( "Log" );
			if ( logElement != null ) {
				if ( logElement.Attribute( "enabled" ) != null ) {
					m.Logging.Enabled = Convert.ToBoolean( logElement.Attribute( "enabled" ).Value );
				}
				if ( logElement.Attribute( "maxfilesize" ) != null ) {
					m.Logging.MaxFileSize = Convert.ToInt32( logElement.Attribute( "maxfilesize" ).Value );
				}
			}
			//load crash recovery
			m.CrashRecovery = new MinerCrashRecoverySetting();
			XElement recoveryElement = minerElement.Element( "CrashRecovery" );
			if ( recoveryElement != null ) {
				if ( recoveryElement.Attribute( "retries" ) != null ) {
					m.CrashRecovery.Retries = Convert.ToInt32( recoveryElement.Attribute( "retries" ).Value );
				}
				if ( recoveryElement.Attribute( "interval" ) != null ) {
					m.CrashRecovery.Interval = Convert.ToInt32( recoveryElement.Attribute( "interval" ).Value );
				}
				if ( recoveryElement.Attribute( "altminer" ) != null ) {
					m.CrashRecovery.AlternateMiner = recoveryElement.Attribute( "altminer" ).Value;
				}
			}
			//load process
			m.Process = new MinerProcessSetting();
			XElement procElement = minerElement.Element( "Process" );
			if ( procElement != null ) {
				if ( procElement.Attribute( "priority" ) != null ) {
					m.Process.Priority = Convert.ToInt16( procElement.Attribute( "priority" ).Value );
				}
				if ( procElement.Attribute( "loadprofile" ) != null ) {
					m.Process.LoadUserProfile = Convert.ToBoolean( procElement.Attribute( "loadprofile" ).Value );
				}
				if ( procElement.Attribute( "windomain" ) != null ) {
					m.Process.WinDomain = procElement.Attribute( "windomain" ).Value;
				}
				if ( procElement.Attribute( "winusername" ) != null ) {
					m.Process.WinUserName = procElement.Attribute( "winusername" ).Value;
				}
				if ( procElement.Attribute( "winpassword" ) != null ) {
					m.Process.WinPassword = procElement.Attribute( "winpassword" ).Value;
				}
				if ( procElement.Attribute( "cpuaffinity" ) != null ) {
					m.Process.CPUAffinity = ( procElement.Attribute( "cpuaffinity" ).Value ?? String.Empty ).Trim();
				}
			}
			return m;
		}

		#endregion

		#region " Saving "

		/// <summary>
		/// Saves all current settings to the XML file.
		/// </summary>
		public void Save( string fileName ) {
			FileInfo fi = new FileInfo( fileName );
			if ( fi.Directory.Exists == false ) {
				Directory.CreateDirectory( fi.DirectoryName );
			}
			XDocument xdoc = new XDocument();
			XElement root = new XElement( "BitMoose" );
			root.Add( new XElement( "DebugLogging", this.DebugLogging ) );
			root.Add( new XElement( "DefaultUserName", this.DefaultUserName ) );
			root.Add( new XElement( "DefaultPassword", this.DefaultPassword ) );
			root.Add( new XElement( "DefaultMinerOption", this.DefaultMinerOption ) );
			root.Add( new XElement( "DefaultPool", this.DefaultPool ) );
			SaveMiners( root );
			SaveMinerOptions( root );
			SaveMinerPools( root );
			xdoc.Add( root );
			xdoc.Save( fileName );
		}

		private void SaveMiners( XElement root ) {
			XElement miners = new XElement( "Miners" );
			foreach ( var m in this.Miners ) {
				XElement miner = new XElement( "Miner" );
				AddAttribute( miner, "name", m.Key );
				AddAttribute( miner, "enabled", m.Value.Enabled );
				AddAttribute( miner, "pool", m.Value.Pool, true );
				AddAttribute( miner, "option", m.Value.OptionKey );
				AddAttribute( miner, "arguments", m.Value.Arguments, true );
				AddAttribute( miner, "host", m.Value.Host, true );
				AddAttribute( miner, "port", m.Value.Port, true );
				AddAttribute( miner, "username", m.Value.UserName, true );
				AddAttribute( miner, "password", m.Value.Password, true );
				AddAttribute( miner, "netupdate", m.Value.Enabled );
				XElement log = new XElement( "Log" );
				AddAttribute( log, "enabled", m.Value.Logging.Enabled );
				AddAttribute( log, "maxfilesize", m.Value.Logging.MaxFileSize );
				miner.Add( log );
				XElement crash = new XElement( "CrashRecovery" );
				AddAttribute( crash, "retries", m.Value.CrashRecovery.Retries );
				AddAttribute( crash, "interval", m.Value.CrashRecovery.Interval );
				AddAttribute( crash, "altminer", m.Value.CrashRecovery.AlternateMiner, true );
				miner.Add( crash );
				XElement process = new XElement( "Process" );
				AddAttribute( process, "priority", m.Value.Process.Priority );
				AddAttribute( process, "cpuaffinity", m.Value.Process.CPUAffinity, true );
				AddAttribute( process, "loadprofile", m.Value.Process.LoadUserProfile );
				AddAttribute( process, "windomain", m.Value.Process.WinDomain, true );
				AddAttribute( process, "winusername", m.Value.Process.WinUserName, true );
				AddAttribute( process, "winpassword", m.Value.Process.WinPassword, true );
				miner.Add( process );
				miners.Add( miner );
			}
			root.Add( miners );
		}

		private void SaveMinerOptions( XElement root ) {
			XElement options = new XElement( "MinerOptions" );
			foreach ( var mo in this.MinerOptions ) {
				XElement option = new XElement( "MinerOption" );
				AddAttribute( option, "name", mo.Key );
				AddAttribute( option, "path", mo.Value.Path );
				AddAttribute( option, "argumentformat", mo.Value.ArgumentFormat );
				AddAttribute( option, "arguments", mo.Value.Arguments );
				options.Add( option );
			}
			root.Add( options );
		}

		private void SaveMinerPools( XElement root ) {
			XElement pools = new XElement( "Pools" );
			foreach ( var p in this.Pools ) {
				XElement pool = new XElement( "Pool" );
				AddAttribute( pool, "name", p.Name );
				AddAttribute( pool, "host", p.Host );
				AddAttribute( pool, "port", p.Port );
				AddAttribute( pool, "website", p.Website, true );
				pools.Add( pool );
			}
			root.Add( pools );
		}

		private void AddAttribute( XElement element, string name, object value ) {
			AddAttribute( element, name, value, false );
		}

		private void AddAttribute( XElement element, string name, object value, bool skipIfNull ) {
			if ( skipIfNull ) {
				if ( value != null ) {
					element.Add( new XAttribute( name, value ) );
				}
			} else {
				element.Add( new XAttribute( name, value ) );
			}
		}

		#endregion

		#endregion

	}
}
