﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Security;
using System.Security.Cryptography;
using System.Management;
using System.Net;
using System.Net.NetworkInformation;
using System.Diagnostics;
using System.IO;

namespace BitMoose.Core {

	public static class Utilities {

		/// <summary>
		/// Returns a secure string object from a non-secure string.
		/// </summary>
		public static SecureString ToSecureString( this string input ) {
			if ( input != null ) {
				SecureString ss = new SecureString();
				for ( int x = 0; x < input.Length; x++ ) {
					ss.AppendChar( input[x] );
				}
				return ss;
			}
			return null;
		}

		public static bool IsNullOrWhiteSpace( this string input ) {
			if ( String.IsNullOrEmpty( input ) == false && input.Trim().Length > 0 ) {
				return false;
			}
			return true;
		}

		public static string CleanString( string input ) {
			if ( input != null ) {
				StringBuilder sb = new StringBuilder();
				for ( int x = 0; x < input.Length; x++ ) {
					if ( Char.IsLetterOrDigit( input[x] ) || Char.IsSeparator( input[x] ) || Char.IsPunctuation( input[x] ) || Char.IsWhiteSpace( input[x] ) ) {
						sb.Append( input[x] );
					}
				}
				return sb.ToString();
			}
			return input;
		}

		public static string GetFQDN() {
			string domainName = IPGlobalProperties.GetIPGlobalProperties().DomainName;
			string hostName = Dns.GetHostName();
			string fqdn = "";
			if ( !hostName.Contains( domainName ) )
				fqdn = hostName + "." + domainName;
			else
				fqdn = hostName;

			return fqdn;
		}

		public static ProcessPriorityClass ToProcessPriority( int priority ) {
			if ( priority < 0 ) {
				return ProcessPriorityClass.BelowNormal;
			} else if ( priority > 0 ) {
				return ProcessPriorityClass.AboveNormal;
			}
			return ProcessPriorityClass.Normal;
		}

		/// <summary>
		/// Parses values by the given argument format
		/// </summary>
		public static string ParseArguments( string argFormat, string host, string userName, string password, short port, string additionalArgs ) {
			string args = argFormat;
			args = args.Replace( "{host}", host );
			args = args.Replace( "{username}", userName );
			args = args.Replace( "{password}", password );
			args = args.Replace( "{port}", port.ToString() );
			if ( args.Contains( "{combinedhost}" ) ) {
				if ( String.IsNullOrEmpty( host ) == false && userName != null ) {
					string authBlock = userName;
					if ( password != null ) {
						authBlock += ":" + password;
					}
					if ( host.Contains( "://" ) ) {
						args = args.Replace( "{combinedhost}", host.Insert( host.IndexOf( "://" ), authBlock + '@' ) );
					} else {
						args = args.Replace( "{combinedhost}", "http://" + authBlock + '@' + host );
					}
				} else {
					args = args.Replace( "{combinedhost}", host );
				}
			}
			args = args.Replace( "{args}", additionalArgs );
			return args;
		}

		/// <summary>
		/// Creates and starting a miner process
		/// </summary>
		public static Process CreateMiningProcess( string minerName, string path, string args, string winDomain, string winUserName, SecureString winPassword, bool loadUserProfile ) {
			ProcessStartInfo psi = new ProcessStartInfo( path, args );
			psi.UseShellExecute = false;
			psi.WorkingDirectory = Path.GetDirectoryName( Path.GetFullPath( path ) );
			psi.Domain = winDomain;
			psi.UserName = winUserName;
			psi.Password = winPassword; ;
			psi.LoadUserProfile = loadUserProfile;
			//psi.Verb = "runas";
			psi.EnvironmentVariables.Add( "BitMoose.MinerName", minerName );
			Process proc = new Process();
			proc.StartInfo = psi;
			return proc;
		}

		public static void StartMiningProcess( Process proc, ProcessPriorityClass priority, long? cpuAffinity ) {
			proc.Start();
			proc.PriorityClass = priority;
			if ( cpuAffinity.HasValue ) {
				proc.ProcessorAffinity = new IntPtr( cpuAffinity.Value );
			}
		}

		/// <summary>
		/// Converts a string representation of a range of cpus into a cpu affinity that can be used with an IntPtr
		/// </summary>
		public static long? ParseCPUAffinity( string cpus ) {
			if ( String.IsNullOrEmpty( cpus ) == false ) {
				if ( cpus.Contains( ',' ) ) { //parse a comma-seperated list of cpus, e.g. "1,2,4,6"
					long affinity = 0;
					string[] temp = cpus.Split( new string[] { "," }, StringSplitOptions.RemoveEmptyEntries );
					for ( int x = 0; x < temp.Length; x++ ) {
						int cpuVal = 0;
						if ( int.TryParse( temp[x], out cpuVal ) && cpuVal > 0 ) {
							long tempaffinity = 1 << cpuVal - 1;
							affinity |= tempaffinity;
						}
					}
					if ( affinity > 0 ) {
						return affinity;
					}
				} else if ( cpus.Contains( '-' ) ) { //parse a range, e.g. "3-4" or "1-8"
					int cpuValLow = 0;
					int cpuValHigh = 0;
					long affinity = 0;
					int dashIndex = cpus.IndexOf( '-' );
					if ( int.TryParse( cpus.Substring( 0, dashIndex ), out cpuValLow ) ) {
						if ( int.TryParse( cpus.Substring( dashIndex + 1, cpus.Length - dashIndex - 1 ), out cpuValHigh ) ) {
							for ( int x = cpuValLow; x <= cpuValHigh; x++ ) {
								affinity |= ( 1 << x - 1 );
							}
							return affinity;
						}
					}
				} else { //parse a single cpu value, e.g. "5"
					int cpuVal = 0;
					if ( int.TryParse( cpus, out cpuVal ) && cpuVal > 0 ) {
						long affinity = 1 << cpuVal - 1;
						return affinity;
					}
				}
			}
			return null;
		}

	}

}
