﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using BitMoose.Core;
using BitMoose.Core.Settings;
using BitMoose.Core.Communication;

namespace BitMoose.App.GUI.Controls {

	public delegate void UpdateStatusHandler( string message );

	public partial class MinerControl : UserControl {

		public event EventHandler HasChangesUpdated;

		private bool Loading = true;

		private bool m_HasChanges = false;

		#region " Properties "

		public string MinerName { get; set; }

		public BitMooseSettings Settings { get; set; }

		public bool HasChanges {
			get { return m_HasChanges; }
			set {
				bool changed = ( value != m_HasChanges );
				m_HasChanges = value;
				if ( Loading == false && changed && HasChangesUpdated != null ) {
					HasChangesUpdated( this, new EventArgs() );
				}
			}
		}

		#endregion

		#region " Constructor(s) "

		public MinerControl() : this( null, null ) { }

		public MinerControl( string minerName, BitMooseSettings settings ) {
			InitializeComponent();
			this.MinerName = minerName;
			this.Settings = settings;
		}

		#endregion

		#region " UI Events "

		private void MinerControl_Load( object sender, EventArgs e ) {
			Loading = true;
			PopulateDefaults();
			PopulateFromSettings();
			this.HasChanges = false;
			Loading = false;
			this.tabControlMain.TabPages.Remove( this.tabPageRecovery );
		}

		private void comboBoxPool_SelectedIndexChanged( object sender, EventArgs e ) {
			PopulateFromPool( this.comboBoxPool.SelectedItem.ToString() );
			this.HasChanges = true;
		}

		private void checkBoxProcessUseDiffCredentials_CheckedChanged( object sender, EventArgs e ) {
			this.textBoxProcessDomain.Enabled =
				this.textBoxProcessPassword.Enabled =
				this.textBoxProcessUserName.Enabled = this.checkBoxProcessUseDiffCredentials.Checked;
			this.HasChanges = true;
		}

		private void comboBoxMinerOption_SelectedIndexChanged( object sender, EventArgs e ) {
			string option = this.comboBoxMinerOption.SelectedItem.ToString();
			this.textBoxMinerArguments.Text = this.Settings.MinerOptions[option].Arguments ?? String.Empty;
			this.HasChanges = true;
		}

		private void linkLabelWebsite_LinkClicked( object sender, LinkLabelLinkClickedEventArgs e ) {
			string uri = this.linkLabelWebsite.Text ?? String.Empty;
			if ( uri.StartsWith( "http" ) && Uri.IsWellFormedUriString( uri, UriKind.Absolute ) ) {
				System.Diagnostics.Process.Start( uri );
			}
		}

		private void numericUp_ValueChanged( object sender, EventArgs e ) {
			this.HasChanges = true;
		}

		private void textBox_TextChanged( object sender, EventArgs e ) {
			this.HasChanges = true;
		}

		private void checkBox_CheckedChanged( object sender, EventArgs e ) {
			this.HasChanges = true;
		}

		private void comboBox_SelectedIndexChanged( object sender, EventArgs e ) {
			this.HasChanges = true;
		}

		private void buttonTestLaunch_Click( object sender, EventArgs e ) {
			LaunchInConsole();
		}

		#endregion

		#region " Methods "

		/// <summary>
		/// Populates default values into the form.
		/// </summary>
		private void PopulateDefaults() {
			HardwareDetector hw = new HardwareDetector();
			hw.Detect();
			this.comboBoxPool.Items.Clear();
			for ( int x = 0; x < this.Settings.Pools.Length; x++ ) {
				this.comboBoxPool.Items.Add( this.Settings.Pools[x].Name );
			}
			this.comboBoxPool.SelectedIndex = -1;
			this.comboBoxMinerOption.Items.Clear();
			foreach ( var kvp in this.Settings.MinerOptions ) {
				this.comboBoxMinerOption.Items.Add( kvp.Key );
			}
			this.comboBoxMinerOption.SelectedItem = this.Settings.DefaultMinerOption;
			comboBoxMinerOption_SelectedIndexChanged( null, null );
			UpdateFallbackMiners();
			this.comboBoxMinerCPUAffinity.Items.Clear();
			this.comboBoxMinerCPUAffinity.Items.Add( "[All Cores]" );
			for ( int x = 1; x <= hw.CPUPhysicalCores; x++ ) {
				this.comboBoxMinerCPUAffinity.Items.Add( x );
			}
			if ( hw.CPUPhysicalCores >= 2 ) {
				for ( int x = 1; x <= hw.CPUPhysicalCores; x++ ) {
					if ( x + 1 <= hw.CPUPhysicalCores ) {
						this.comboBoxMinerCPUAffinity.Items.Add( String.Format( "{0}-{1}", x, x + 1 ) );
						x++;
					}
				}
			}
			this.comboBoxMinerCPUAffinity.SelectedIndex = 0;
			this.comboBoxMinerPriority.SelectedIndex = 1;
			this.textBoxUserName.Text = this.Settings.DefaultUserName;
			this.textBoxPassword.Text = this.Settings.DefaultPassword;
			this.checkBoxLogEnabled.Checked = true;
			this.numericUpDownLogMaxSize.Value = 1048576;
			this.numericUpDownCrashInterval.Value = 60;
			this.numericUpDownCrashRetries.Value = 3;
			this.checkBoxProcessUseDiffCredentials.Checked = false;
			this.textBoxProcessDomain.Text = Utilities.GetFQDN();
			this.textBoxProcessUserName.Text = String.Empty;
			this.textBoxProcessPassword.Text = String.Empty;
			this.checkBoxPauseStatusUpdates.Checked = false;
		}

		/// <summary>
		/// Populates the UI from values in the settings
		/// </summary>
		private void PopulateFromSettings() {
			MinerSetting ms = this.Settings.Miners[this.MinerName];
			PopulateFromPool( ms.Pool );
			if ( String.IsNullOrEmpty( ms.Host ) == false ) {
				this.textBoxHost.Text = ms.Host;
			}
			if ( ms.Port.HasValue ) {
				this.numericUpDownPort.Value = ms.Port.Value;
			}
			this.textBoxUserName.Text = ms.UserName;
			this.textBoxPassword.Text = ms.Password;
			this.comboBoxMinerOption.SelectedItem = ms.OptionKey;
			this.textBoxMinerArguments.Text = ms.Arguments;
			if ( String.IsNullOrEmpty( ms.Process.CPUAffinity ) == false ) {
				this.comboBoxMinerCPUAffinity.Text = ms.Process.CPUAffinity;
			} else {
				this.comboBoxMinerCPUAffinity.SelectedIndex = 0;
			}
			this.comboBoxMinerPriority.SelectedIndex = ms.Process.Priority + 1;
			this.checkBoxLogEnabled.Checked = ms.Logging.Enabled;
			this.numericUpDownLogMaxSize.Value = ms.Logging.MaxFileSize;
			this.checkBoxProcessUseDiffCredentials.Checked = !( ms.Process.WinUserName.IsNullOrWhiteSpace() && ms.Process.WinPassword.IsNullOrWhiteSpace() );
			this.textBoxProcessDomain.Text = ms.Process.WinDomain;
			this.textBoxProcessUserName.Text = ms.Process.WinUserName;
			this.textBoxProcessPassword.Text = ms.Process.WinPassword;
			this.numericUpDownCrashRetries.Value = ms.CrashRecovery.Retries;
			this.numericUpDownCrashInterval.Value = ms.CrashRecovery.Interval;
			this.comboBoxCrashFallback.SelectedItem = ms.CrashRecovery.AlternateMiner;
		}

		/// <summary>
		/// Populates the UI with pool values
		/// </summary>
		private void PopulateFromPool( string poolName ) {
			MiningPool pool = null;
			if ( poolName != null ) {
				pool = this.Settings.Pools.FirstOrDefault( a => a.Name == poolName );
			}
			if ( pool != null ) {
				this.comboBoxPool.SelectedItem = pool.Name;
				this.textBoxHost.Text = pool.Host;
				this.numericUpDownPort.Value = pool.Port;
				if ( pool.Website.IsNullOrWhiteSpace() ) {
					this.linkLabelWebsite.Text = "No Website Available";
				} else {
					this.linkLabelWebsite.Text = pool.Website;
				}
			}
		}

		/// <summary>
		/// Updates the miner status window with a message
		/// </summary>
		public void UpdateStatus( string message ) {
			if ( this.checkBoxPauseStatusUpdates.Checked == false ) {
				int caretMaintain = -1;
				if ( this.textBoxStatus.Focused ) {
					caretMaintain = this.textBoxStatus.SelectionStart;
				}
				if ( message.Length + this.textBoxStatus.Text.Length > this.textBoxStatus.MaxLength ) {
					this.textBoxStatus.ResetText();
				}
				this.textBoxStatus.Text = message + Environment.NewLine + this.textBoxStatus.Text;
				if ( caretMaintain > -1 ) {
					this.textBoxStatus.SelectionStart = caretMaintain;
					this.textBoxStatus.ScrollToCaret();
				}
			}
		}

		/// <summary>
		/// Updates the list of fallback miners
		/// </summary>
		public void UpdateFallbackMiners() {
			string selected = this.comboBoxCrashFallback.SelectedItem as string;
			this.comboBoxCrashFallback.Items.Clear();
			foreach ( var kvp in this.Settings.Miners ) {
				if ( kvp.Key != this.MinerName ) {
					this.comboBoxCrashFallback.Items.Add( kvp.Key );
				}
			}
			if ( selected != null ) {
				if ( this.comboBoxCrashFallback.Items.Contains( selected ) ) {
					this.comboBoxCrashFallback.SelectedItem = selected;
				} else {
					this.comboBoxCrashFallback.SelectedIndex = -1;
					this.HasChanges = true;
				}
			} else {
				this.comboBoxCrashFallback.SelectedIndex = -1;
			}
		}

		/// <summary>
		/// Saves the contents of the form to the settings held in memory
		/// </summary>
		public void SaveToSettings() {
			MinerSetting ms = Settings.Miners[this.MinerName];
			ms.Arguments = this.textBoxMinerArguments.Text;
			ms.Host = this.textBoxHost.Text;
			ms.OptionKey = this.comboBoxMinerOption.SelectedItem.ToString();
			ms.UserName = this.textBoxUserName.Text;
			ms.Password = this.textBoxPassword.Text;
			ms.Pool = this.comboBoxPool.SelectedItem.ToString();
			ms.Port = (short)this.numericUpDownPort.Value;
			ms.CrashRecovery.AlternateMiner = ( this.comboBoxPool.SelectedItem ?? String.Empty ).ToString();
			ms.CrashRecovery.Interval = (int)this.numericUpDownCrashInterval.Value;
			ms.CrashRecovery.Retries = (int)this.numericUpDownCrashRetries.Value;
			ms.Logging.Enabled = this.checkBoxLogEnabled.Checked;
			ms.Logging.MaxFileSize = (int)this.numericUpDownLogMaxSize.Value;
			ms.Process.CPUAffinity = this.comboBoxMinerCPUAffinity.Text;
			ms.Process.Priority = (short)( this.comboBoxMinerPriority.SelectedIndex - 1 );
			if ( this.checkBoxProcessUseDiffCredentials.Checked ) {
				ms.Process.WinDomain = this.textBoxProcessDomain.Text;
				ms.Process.WinUserName = this.textBoxProcessUserName.Text;
				ms.Process.WinPassword = this.textBoxProcessPassword.Text;
			} else {
				ms.Process.WinDomain = null;
				ms.Process.WinUserName = null;
				ms.Process.WinPassword = null;
			}
			this.HasChanges = false;
		}

		public void LaunchInConsole() {
			MinerOption option = Settings.MinerOptions[this.comboBoxMinerOption.SelectedItem.ToString()];
			string args = Utilities.ParseArguments( option.ArgumentFormat, this.textBoxHost.Text, this.textBoxUserName.Text, this.textBoxPassword.Text, (short)this.numericUpDownPort.Value, this.textBoxMinerArguments.Text );
			string winDomain = null;
			string winUsername = null;
			string winPassword = null;
			if ( this.checkBoxProcessUseDiffCredentials.Checked ) {
				winDomain = this.textBoxProcessDomain.Text;
				winUsername = this.textBoxProcessUserName.Text;
				winPassword = this.textBoxProcessPassword.Text;
			}
			ProcessPriorityClass priority = Utilities.ToProcessPriority( this.comboBoxMinerPriority.SelectedIndex - 1 );
			long? cpuAffinity = Utilities.ParseCPUAffinity( this.comboBoxMinerCPUAffinity.Text );
			Process proc = Utilities.CreateMiningProcess( this.MinerName, option.Path, args, winDomain, winUsername, winPassword.ToSecureString(), false );
			Utilities.StartMiningProcess( proc, priority, cpuAffinity );
		}

		#endregion

	}

}
