﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.IO;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics;
using System.Threading;
using System.ServiceProcess;
using BitMoose.Core;
using BitMoose.Core.Communication;
using BitMoose.Core.Settings;
using BitMoose.App.GUI.Controls;

namespace BitMoose.App.GUI {

	public partial class MainForm : Form {

		private Client ServiceClient = null;
		private BitMooseSettings Settings = null;
		private Dictionary<string, MinerControl> MinerControls = new Dictionary<string, MinerControl>();
		private Thread StatusThread = null;

		#region " Constructor(s) "

		public MainForm() {
			InitializeComponent();
		}

		#endregion

		#region " UI Events "

		private void MainForm_Load( object sender, EventArgs e ) {
			string fileName = Path.Combine( Path.GetDirectoryName( Process.GetCurrentProcess().MainModule.FileName ), BitMooseSettings.SettingsFileName );
			if ( LoadSettings( fileName ) == false ) {
				this.Close();
				return;
			}
			this.ServiceClient = new Client();
			this.ServiceClient.ClientReceivedMessage += new ClientReceivedMessageHandler( ServiceClient_ClientReceivedMessage );
			this.ServiceClient.Start();
			StatusThread = new Thread( new ThreadStart( StatusCheckLoop ) );
			StatusThread.Start();
			PopulateUI();
		}

		private void MainForm_FormClosing( object sender, FormClosingEventArgs e ) {
			StatusThread.Abort();
			if ( this.ServiceClient != null ) {
				this.ServiceClient.Stop();
			}
		}

		private void ServiceClient_ClientReceivedMessage( string sender, string message ) {
			if ( MinerControls.ContainsKey( sender ) ) {
				if ( MinerControls[sender].InvokeRequired ) {
					MinerControls[sender].Invoke( new UpdateStatusHandler( MinerControls[sender].UpdateStatus ), new object[] { message } );
				} else {
					MinerControls[sender].UpdateStatus( message );
				}
			}
		}

		private void tabControlMain_Selected( object sender, TabControlEventArgs e ) {
			UpdateMenus();
		}

		private void MinerControl_HasChangesUpdated( object sender, EventArgs e ) {
			UpdateTabs();
		}

		#region " File Menu "

		private void minersToolStripMenuItem_Click( object sender, EventArgs e ) {
			MinerSetting ms = new MinerSetting();
			int count = this.Settings.Miners.Count + 1;
			string name = String.Format( "Miner #{0}", count );
			//find an unused name
			while ( this.Settings.Miners.ContainsKey( name ) ) {
				count++;
				name = String.Format( "Miner #{0}", count );
			}
			ms.Pool = this.Settings.DefaultPool;
			ms.OptionKey = this.Settings.DefaultMinerOption;
			ms.UserName = this.Settings.DefaultUserName;
			ms.Password = this.Settings.DefaultPassword;
			ms.Arguments = this.Settings.MinerOptions[ms.OptionKey].Arguments;
			this.Settings.Miners.Add( name, ms );
			this.tabControlMain.SelectedTab = CreateMinerTab( name, ms.Enabled );
			UpdateMinerControls();
		}

		private void saveConfigurationToolStripMenuItem_Click( object sender, EventArgs e ) {
			Save( Path.Combine( Settings.BitMoosePath, BitMooseSettings.SettingsFileName ) );
		}

		private void saveAsToolStripMenuItem_Click( object sender, EventArgs e ) {
			SaveFileDialog sfd = new SaveFileDialog();
			sfd.FileName = BitMooseSettings.SettingsFileName;
			sfd.InitialDirectory = Settings.BitMoosePath;
			sfd.Filter = "XML File (*.xml)|*.xml|All Files|*";
			if ( sfd.ShowDialog() == DialogResult.OK ) {
				Save( sfd.FileName );
			}
		}

		private void loadToolStripMenuItem_Click( object sender, EventArgs e ) {
			OpenFileDialog ofd = new OpenFileDialog();
			ofd.FileName = BitMooseSettings.SettingsFileName;
			ofd.InitialDirectory = Settings.BitMoosePath;
			ofd.RestoreDirectory = true;
			ofd.Filter = "XML File (*.xml)|*.xml|All Files|*";
			if ( ofd.ShowDialog() == DialogResult.OK ) {
				LoadSettings( ofd.FileName );
				PopulateUI();
			}
		}

		private void exitToolStripMenuItem_Click( object sender, EventArgs e ) {
			bool hasChanges = false;
			foreach ( var mc in this.MinerControls ) {
				if ( mc.Value.HasChanges ) {
					hasChanges = true;
					break;
				}
			}
			if ( hasChanges ) {
				if ( MessageBox.Show( "Save changes before exiting?", "Save Changes", MessageBoxButtons.YesNo, MessageBoxIcon.Question ) == DialogResult.Yes ) {
					Save( Path.Combine( Settings.BitMoosePath, BitMooseSettings.SettingsFileName ) );
				}
			}
			this.Close();
		}

		#endregion

		#region " Services Menu "

		private void stopToolStripMenuItem_Click( object sender, EventArgs e ) {
			BitMooseServiceController.Stop();
		}

		private void startToolStripMenuItem_Click( object sender, EventArgs e ) {
			BitMooseServiceController.Start();
		}

		private void installToolStripMenuItem_Click( object sender, EventArgs e ) {
			BitMooseServiceController.Install( Settings.BitMoosePath );
		}

		private void uninstallToolStripMenuItem_Click( object sender, EventArgs e ) {
			BitMooseServiceController.Uninstall();
		}

		#endregion

		#region " Miner Menu "

		private void enabledToolStripMenuItem_Click( object sender, EventArgs e ) {
			string name = null;
			TabPage page = null;
			MinerSetting ms = GetSelectedMiner( out name, out page );
			if ( ms != null ) {
				this.Settings.Miners[name].Enabled = this.enabledToolStripMenuItem.Checked;
				page.ImageKey = ( this.Settings.Miners[name].Enabled ? "node.png" : "node_fade.png" );
				( (MinerControl)page.Controls[0] ).HasChanges = true;
				this.minerToolStripMenuItem.ShowDropDown();
			}
		}

		private void statusUpdatesToolStripMenuItem_Click( object sender, EventArgs e ) {
			string name = null;
			TabPage page = null;
			MinerSetting ms = GetSelectedMiner( out name, out page );
			if ( ms != null ) {
				this.Settings.Miners[name].StatusUpdates = this.statusUpdatesToolStripMenuItem.Checked;
				( (MinerControl)page.Controls[0] ).HasChanges = true;
				this.minerToolStripMenuItem.ShowDropDown();
			}
		}

		private void deleteToolStripMenuItem_Click( object sender, EventArgs e ) {
			string name = null;
			TabPage page = null;
			MinerSetting ms = GetSelectedMiner( out name, out page );
			if ( ms != null ) {
				if ( MessageBox.Show( String.Format( "Are you sure you want to delete the miner '{0}'?", name ), "Delete Miner?", MessageBoxButtons.YesNo, MessageBoxIcon.Question ) == System.Windows.Forms.DialogResult.Yes ) {
					this.Settings.Miners.Remove( name );
					this.tabControlMain.TabPages.Remove( page );
					this.MinerControls.Remove( name );
					UpdateMenus();
					UpdateMinerControls();
				}
			}
		}

		private void renameToolStripMenuItem_Click( object sender, EventArgs e ) {
			string name = null;
			TabPage page = null;
			MinerSetting ms = GetSelectedMiner( out name, out page );
			if ( ms != null ) {
				RenameForm rf = new RenameForm( name, this.Settings );
				if ( rf.ShowDialog() == DialogResult.OK ) {
					MinerControl mc = ( (MinerControl)page.Controls[0] );
					this.Settings.Miners.Remove( name );
					this.Settings.Miners.Add( rf.MinerName, ms );
					mc.MinerName = rf.MinerName;
					mc.HasChanges = true;
					page.Text = rf.MinerName + '*';
					UpdateMinerControls();
				}
			}
		}

		#endregion

		#region " Settings Menu "

		private void debugLoggingToolStripMenuItem_Click( object sender, EventArgs e ) {
			this.Settings.DebugLogging = this.debugLoggingToolStripMenuItem.Checked;
			this.settingsToolStripMenuItem.ShowDropDown();
		}

		private void defaultsToolStripMenuItem_Click( object sender, EventArgs e ) {
			new DefaultValuesForm( this.Settings ).ShowDialog();
		}

		#endregion

		#region " Help Menu "

		private void donateToolStripMenuItem_Click( object sender, EventArgs e ) {
			new DonateForm().ShowDialog();
		}

		private void informationToolStripMenuItem_Click( object sender, EventArgs e ) {
			new AboutForm().ShowDialog();
		}

		#endregion

		#endregion

		#region " Methods "

		private bool LoadSettings( string fileName ) {
			try {
				this.Settings = new BitMooseSettings();
				this.Settings.Load( fileName );
				return true;
			} catch ( Exception ex ) {
				string message = "There was an error loading the settings file. The message returned was:" + Environment.NewLine + ex.Message;
				message += Environment.NewLine + Environment.NewLine + "Would you like to recreate the settings file?";
				if ( MessageBox.Show( message, "Error Loading Settings", MessageBoxButtons.YesNo, MessageBoxIcon.Error, MessageBoxDefaultButton.Button2 ) == System.Windows.Forms.DialogResult.Yes ) {
					//TODO recreate settings file
					return true;
				}
			}
			return false;
		}

		private void PopulateUI() {
			this.MinerControls.Clear();
			//create tabs
			this.tabControlMain.TabPages.Clear();
			if ( this.Settings.Miners != null ) {
				foreach ( var ms in this.Settings.Miners ) {
					CreateMinerTab( ms.Key, ms.Value.Enabled );
				}
			}
			//ready the UI
			UpdateMenus();
		}

		private TabPage CreateMinerTab( string name, bool enabled ) {
			TabPage page = new TabPage( name );
			MinerControl mc = new MinerControl( name, this.Settings );
			mc.Dock = DockStyle.Fill;
			mc.HasChangesUpdated += new EventHandler( MinerControl_HasChangesUpdated );
			page.Controls.Add( mc );
			MinerControls.Add( name, mc );
			this.tabControlMain.TabPages.Add( page );
			page.ImageKey = ( enabled ? "node.png" : "node_fade.png" );
			return page;
		}

		private void UpdateMenus() {
			//update miner menu
			this.minerToolStripMenuItem.Enabled = ( this.tabControlMain.SelectedIndex >= 0 );
			if ( this.tabControlMain.SelectedIndex >= 0 ) {
				string name = null;
				TabPage page = null;
				MinerSetting ms = GetSelectedMiner( out name, out page );
				if ( ms != null ) {
					this.enabledToolStripMenuItem.Checked = ms.Enabled;
					this.statusUpdatesToolStripMenuItem.Checked = ms.StatusUpdates;
				}
			}
			//update settings menu
			this.debugLoggingToolStripMenuItem.Checked = this.Settings.DebugLogging;
		}

		private void UpdateTabs() {
			for ( int x = 0; x < this.tabControlMain.TabPages.Count; x++ ) {
				MinerControl mc = this.tabControlMain.TabPages[x].Controls[0] as MinerControl;
				if ( mc != null ) {
					this.tabControlMain.TabPages[x].Text = ( mc.HasChanges ? mc.MinerName + '*' : mc.MinerName );
				}
			}
		}

		private void UpdateMinerControls() {
			foreach ( var m in MinerControls ) {
				m.Value.UpdateFallbackMiners();
			}
		}

		private void StatusCheckLoop() {
			while ( true ) {
				ServiceControllerStatus status = ServiceControllerStatus.Stopped;
				if ( BitMooseServiceController.TryGetStatus( out status ) ) {
					if ( this.toolStripContainerMain.InvokeRequired ) {
						this.toolStripContainerMain.Invoke( (MethodInvoker)delegate { UpdateServiceStatus( status ); } );
					} else {
						UpdateServiceStatus( status );
					}
					Thread.Sleep( 1000 );
				} else {
					if ( this.toolStripContainerMain.InvokeRequired ) {
						this.toolStripContainerMain.Invoke( (MethodInvoker)delegate { UpdateServiceStatus( null ); } );
					} else {
						UpdateServiceStatus( status );
					}
					Thread.Sleep( 7500 );
				}
			}
		}

		private void UpdateServiceStatus( ServiceControllerStatus? status ) {
			if ( status.HasValue == false ) {
				this.toolStripLabelServiceStatus.Text = "Service Not Installed";
				this.toolStripLabelServiceStatus.Image = Properties.Resources.server_delete;
				this.stopToolStripMenuItem.Enabled = false;
				this.startToolStripMenuItem.Enabled = false;
				this.installToolStripMenuItem.Enabled = true;
				this.uninstallToolStripMenuItem.Enabled = false;
			} else {
				this.stopToolStripMenuItem.Enabled = ( status.Value == ServiceControllerStatus.Running );
				this.startToolStripMenuItem.Enabled = ( status.Value == ServiceControllerStatus.Stopped || status.Value == ServiceControllerStatus.Paused );
				this.installToolStripMenuItem.Enabled = false;
				this.uninstallToolStripMenuItem.Enabled = true;
				switch ( status.Value ) {
					case ServiceControllerStatus.StopPending: this.toolStripLabelServiceStatus.Text = "Stopping..."; break;
					case ServiceControllerStatus.Stopped: this.toolStripLabelServiceStatus.Text = "Service Stopped"; break;
					case ServiceControllerStatus.ContinuePending: this.toolStripLabelServiceStatus.Text = "Resuming..."; break;
					case ServiceControllerStatus.PausePending: this.toolStripLabelServiceStatus.Text = "Pausing..."; break;
					case ServiceControllerStatus.Paused: this.toolStripLabelServiceStatus.Text = "Service Paused"; break;
					case ServiceControllerStatus.Running: this.toolStripLabelServiceStatus.Text = "Service Running"; break;
					case ServiceControllerStatus.StartPending: this.toolStripLabelServiceStatus.Text = "Starting..."; break;
				}
				switch ( status.Value ) {
					case ServiceControllerStatus.StopPending:
					case ServiceControllerStatus.Stopped:
					case ServiceControllerStatus.PausePending:
					case ServiceControllerStatus.Paused:
						this.toolStripLabelServiceStatus.Image = Properties.Resources.server_delete;
						break;
					case ServiceControllerStatus.ContinuePending:
					case ServiceControllerStatus.StartPending:
						this.toolStripLabelServiceStatus.Image = Properties.Resources.server_connection;
						break;
					case ServiceControllerStatus.Running:
						this.toolStripLabelServiceStatus.Image = Properties.Resources.server_ok;
						break;
				}
			}
		}

		private void Save( string fileName ) {
			foreach ( var m in MinerControls ) {
				m.Value.SaveToSettings();
			}

#if DEBUG
			Settings.Save( fileName );
#else
			try {
				Settings.Save( fileName );
			} catch ( Exception ex ) {
				MessageBox.Show( "There was an error trying to save the file." + Environment.NewLine + ex.Message, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error );
			}
#endif
		}

		private MinerSetting GetSelectedMiner( out string name, out TabPage page ) {
			MinerOption option = null;
			MiningPool pool = null;
			return GetSelectedMiner( out name, out page, out option, out pool );
		}

		private MinerSetting GetSelectedMiner( out string name, out TabPage page, out MinerOption option, out MiningPool pool ) {
			name = null;
			option = null;
			pool = null;
			page = null;
			if ( this.tabControlMain.TabIndex > -1 ) {
				page = this.tabControlMain.TabPages[this.tabControlMain.SelectedIndex];
				name = page.Text.TrimEnd( '*' );
				MinerSetting ms = this.Settings.Miners[name];
				option = this.Settings.MinerOptions[ms.OptionKey];
				if ( String.IsNullOrEmpty( ms.Pool ) == false ) {
					pool = this.Settings.Pools.Where( a => a.Name == ms.Pool ).FirstOrDefault();
				}
				return ms;
			}
			return null;
		}

		#endregion

	}

}
